# frozen_string_literal: true

require 'json-diff'
require 'yaml'
require 'fileutils'
require 'xcodeproj'

class LocalizationLinter
  def initialize(
    verbose = nil,
    clean_after_execution = nil,
    language_configuration_path = nil,
    swift_gen_path = nil,
    derived_data_path = nil,
    resource_path = nil,
    project_path = nil
  )
    @verbose = !verbose.nil? || verbose == 'true' || verbose == true
    @clean_after_execution = !clean_after_execution.nil? || clean_after_execution == 'true' || clean_after_execution == true
    @language_configuration_path = language_configuration_path
    @swift_gen_path = swift_gen_path
    @derived_data_path = derived_data_path
    @resource_path = resource_path
    @project_path = project_path
  end

  def find_differences
    clean_derived_data
    create_derived_data_folder
    log "Validating wording for targets: #{all_targets.join ', '}"
    differences = all_targets.map { |target| check_missing_keys_in_target(target) }.compact.flatten
    clean_derived_data if @clean_after_execution
    differences
  end

  def log(text)
    puts(text) if @verbose
  end

  def check_missing_keys_in_target(target)
    log "Starting wording validation for #{target}"
    main_region = main_region_for_target(target)
    if main_region.nil?
      return {
        filename: 'General',
        message: "Missing main region for target #{target}"
      }
    end

    log "Found main region #{main_region} for #{target}"
    unless known_regions.include? main_region
      return {
        filename: 'General',
        message: "Missing main region #{main_region} in Project"
      }
    end
    known_regions.each { |region| write_l10n_check_file(target, region) }
    main_file_name = temporary_localizable_check_file_path(target, main_region)
    log "Taking #{main_file_name} as reference for #{target}"
    unless File.exist?(main_file_name)
      return {
        filename: 'General',
        message: "Target #{target} doesn't localize for main region #{main_region}"
      }
    end
    log "Found file for main region #{main_region} for #{target}"
    main_file_hash = get_hash(main_file_name)
    errors = known_regions.map do |region|
      begin
        check_missing_key(target, region, main_region, main_file_hash)
      rescue CheckMissingKeyNotNeeded
        next
      end
    end.compact
    errors.flatten.compact unless errors.empty?
  end

  ## Tools

  class CheckMissingKeyNotNeeded < StandardError
  end

  class IsMainRegion < CheckMissingKeyNotNeeded
  end

  class MissingFileForRegion < CheckMissingKeyNotNeeded
    def initialize(msg="")
      super
    end
  end

  class RegionHashIdenticalAsMainOne < CheckMissingKeyNotNeeded
    def initialize(msg="")
      super
    end
  end

  class RegionHasNoDiffError < CheckMissingKeyNotNeeded
    def initialize(msg="")
      super
    end
  end

  def check_missing_key(target, region, main_region, main_file_hash)
    raise IsMainRegion if region == main_region
    message = "#{target}_#{region}"

    file_name = temporary_localizable_check_file_path(target, region)
    raise MissingFileForRegion, message unless File.exist?(file_name)
    file_hash = get_hash(file_name)
    raise RegionHashIdenticalAsMainOne, message unless file_hash != main_file_hash

    log "Computing diff for #{region} with file: #{file_name}"
    previous_diff = {}
    previous_previous_diff = {}
    diffs = JsonDiff.diff(file_hash, main_file_hash)
    last_diff = diffs.last
    current_errors = []
    diffs.each do |diff|
      current_errors << handle_previous_diff(diff, previous_diff, previous_previous_diff)
      current_errors << handle_last_diff(diff, previous_diff) if diff == last_diff
      previous_previous_diff = previous_diff
      previous_diff = diff
    end
    if current_errors.empty?
      log "Did not found any diff for #{region}"
    else
      log "Found #{current_errors.size} diffs:#{current_errors.join "\n"}"
    end
    raise RegionHasNoDiffError, message if current_errors.compact.empty?

    current_errors.compact.map do |error|
      parse_error_with_file_and_line(error, localizable_directory(target, region))
    end
  end

  def parse_error_with_file_and_line(error, directory)
    final_key = error[:key].gsub('/L10n/', '').split('/').first
    line_vs_file = find_line(final_key, directory)
    {
      filename: line_vs_file[1],
      key: final_key,
      line: line_vs_file[0],
      message: error[:message]
    }
  end

  def find_line(key, directory)
    result = [0, '']
    Dir.foreach(directory) do |filename|
      next if (filename == '.') || (filename == '..') || (result[0] != 0)

      global = "#{directory}/#{filename}"
      result = [0, global] unless result != [0, '']
      File.foreach(global).with_index do |line, index|
        next if result[0] != 0

        result = [index + 1, global] if line.include? key
      end
    end
    result
  end

  def write_l10n_check_file(target, language)
    unless File.directory?(localizable_directory(target, language))
      log "Target #{target} doesn't localize for region #{language}"
      return
    end
    File.open(swiftgen_check_config_path, 'w') { |file| file.write(swiftgen_check_config(target, language).to_yaml) }
    system("#{swift_gen_path} config run --config #{swiftgen_check_config_path}")
  end

  def get_hash(name)
    trim = File.read(name).gsub(/[[:space:]]/, '').gsub(',}', '}')
    JSON.parse(trim)
  end

  def main_region_for_target(target)
    path = language_configuration_path
    unless File.exist?(path)
      log "Language configuration file not found, selecting default region 'en' for #{target}"
      return 'en'
    end
    languages = YAML.safe_load(File.read(path)) || {}
    unless languages[target]
      log "Language configuration file not found, selecting default region 'en' for #{target}"
      return 'en'
    end

    languages[target]
  end

  # Errors

  def handle_previous_diff(diff, previous_diff, previous_previous_diff)
    case previous_diff['op']
    when 'remove'
      useless_key(previous_diff) if previous_diff['path'] != diff['path']
    when 'add'
      if previous_diff['path'] != diff['path']
        if (previous_previous_diff != {}) && (previous_previous_diff['path'] != previous_diff['path'])
          missing_key(previous_diff)
        else
          bad_parameter(previous_diff)
        end
      else
        bad_parameter(previous_diff)
      end
    end
  end

  def handle_last_diff(diff, previous_diff)
    if previous_diff['path'] != diff['path']
      if diff['op'] == 'remove'
        useless_key(diff)
      else
        missing_key(diff)
      end
    else
      bad_parameter(diff)
    end
  end

  def useless_key(diff)
    {
      key: diff['path'],
      message: "useless key: #{diff['path']}"
    }
  end

  def missing_key(diff)
    {
      key: diff['path'],
      message: "missing key: #{diff['path']}"
    }
  end

  def bad_parameter(diff)
    {
      key: diff['path'],
      message: "wrong parameters in key: #{diff['path']}/#{diff['value']}"
    }
  end

  ## Paths

  def localizable_directory(target, language)
    "#{resource_path}/#{target}/#{language}.lproj"
  end

  def swiftgen_check_config_path
    "#{derived_data_path}/swiftgen_check_config.yml"
  end

  def temporary_localizable_check_file_path(target, region)
    "#{derived_data_path}/#{target}_#{region}"
  end

  def derived_data_path
    @derived_data_path || 'localization_linter_derived_data'
  end

  def resource_path
    @resource_path || 'Resources'
  end

  def swift_gen_path
    @swift_gen_path || 'Pods/SwiftGen/bin/swiftgen'
  end

  def language_configuration_path
    @language_configuration_path || 'Configuration/Defaults/all_target_main_language.yml'
  end

  def stencil_path
    "#{File.expand_path(File.dirname(__FILE__))}/swiftgen_localizables_check.stencil"
  end

  def upper_path(path)
    return "../#{path}" unless path.start_with?('/')
    return path
  end

  ## Derived data

  def create_derived_data_folder
    Dir.mkdir derived_data_path
  end

  def clean_derived_data
    FileUtils.rm_rf(derived_data_path)
  end

  ## config

  def swiftgen_check_config(target, language)
    {
      'strings' => {
        'inputs' => [
          "#{upper_path(localizable_directory(target, language))}"
        ],
        'outputs' => [{
          'templatePath' => stencil_path,
          'output' => "#{upper_path(temporary_localizable_check_file_path(target, language))}",
          'params' => {
            'publicAccess' => true
          }
        }]
      }
    }
  end

  ## project

  def project_path
    @project_path || Dir.glob('*.xcodeproj').first
  end

  def project
    Xcodeproj::Project.open(project_path)
  end

  def all_targets
    project
      .targets
      .filter { |t| t.product_type == 'com.apple.product-type.application' }
      .map(&:name)
  end

  def known_regions
    project.root_object.known_regions
  end
end
